package com.perforce.workshop.tjuricek.p4oauth.test_server;

import com.github.jknack.handlebars.Handlebars;
import com.github.jknack.handlebars.Template;
import org.apache.commons.io.IOUtils;
import org.eclipse.jetty.server.Request;
import org.eclipse.jetty.server.handler.AbstractHandler;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.net.InetAddress;
import java.net.UnknownHostException;

/**
 * If we have an .hbs file with the indicated path name, we'll load that
 * and apply a consistent "configuration" context to it.
 */
public class HandlebarsHelper extends AbstractHandler {

    private Handlebars handlebars = new Handlebars();
    private TemplateContext templateContext;

    public HandlebarsHelper(int port) {
        this.templateContext = new TemplateContext();
        templateContext.setPort(port);
    }

    @Override
    public void handle(String target, Request baseRequest, HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {

        if (!baseRequest.getMethod().equals("GET")) {
            return;
        }

        String path = request.getPathInfo() + ".hbs";

        InputStream hbsStream =
                ClassLoader.getSystemResourceAsStream("META-INF/resources" + path);

        if (hbsStream != null) {
            baseRequest.setHandled(true);
            Template template = handlebars.compileInline(IOUtils.toString(hbsStream));

            OutputStreamWriter writer = new OutputStreamWriter(response.getOutputStream());
            try {
                response.addHeader("Access-Control-Allow-Origin",
                        "https://" + templateContext.getHostname() + ":" + templateContext.getP4oAuthPort());
                response.addHeader("Access-Control-Allow-Methods",
                        "POST");

                if (request.getParameter("access_token") != null) {
                    response.addHeader("Authorization", "Bearer " + request.getParameter("access_token"));
                }

                response.setContentType("text/html");
                template.apply(templateContext, writer);
            } finally {
                IOUtils.closeQuietly(writer);
            }
        }
    }

    public TemplateContext getTemplateContext() {
        return templateContext;
    }

    public void setTemplateContext(TemplateContext templateContext) {
        this.templateContext = templateContext;
    }

    public static class TemplateContext {
        private String hostname = null;
        private String p4oAuthHostname;
        private int p4oAuthPort = 8443;
        private String login = "jdoe";
        private String redirectUri;
        private int port = 8444;

        TemplateContext() {
            hostname = Conventions.getHostname();

            if (hostname != null) {
                updateHostname();
            }
        }

        public String getP4oAuthHostname() {
            return p4oAuthHostname;
        }

        public void setP4oAuthHostname(String p4oAuthHostname) {
            this.p4oAuthHostname = p4oAuthHostname;
        }

        public int getP4oAuthPort() {
            return p4oAuthPort;
        }

        public void setP4oAuthPort(int p4oAuthPort) {
            this.p4oAuthPort = p4oAuthPort;
        }

        public String getLogin() {
            return login;
        }

        public void setLogin(String login) {
            this.login = login;
        }

        public String getRedirectUri() {
            return redirectUri;
        }

        public void setRedirectUri(String redirectUri) {
            this.redirectUri = redirectUri;
        }

        public String getHostname() {
            return hostname;
        }

        public void setHostname(String hostname) {
            this.hostname = hostname;
            updateHostname();
        }

        private void updateHostname() {
            if (hostname != null) {
                p4oAuthHostname = hostname;
                redirectUri = Conventions.getRedirectUri(hostname, port);
            }
        }

        public int getPort() {
            return port;
        }

        public void setPort(int port) {
            this.port = port;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            TemplateContext that = (TemplateContext) o;

            if (p4oAuthPort != that.p4oAuthPort) return false;
            if (port != that.port) return false;
            if (hostname != null ? !hostname.equals(that.hostname) : that.hostname != null)
                return false;
            if (login != null ? !login.equals(that.login) : that.login != null)
                return false;
            if (p4oAuthHostname != null ? !p4oAuthHostname.equals(that.p4oAuthHostname) : that.p4oAuthHostname != null)
                return false;
            if (redirectUri != null ? !redirectUri.equals(that.redirectUri) : that.redirectUri != null)
                return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result = hostname != null ? hostname.hashCode() : 0;
            result = 31 * result + (p4oAuthHostname != null ? p4oAuthHostname.hashCode() : 0);
            result = 31 * result + p4oAuthPort;
            result = 31 * result + (login != null ? login.hashCode() : 0);
            result = 31 * result + (redirectUri != null ? redirectUri.hashCode() : 0);
            result = 31 * result + port;
            return result;
        }

        @Override
        public String toString() {
            return "TemplateContext{" +
                    "p4oAuthHostname='" + p4oAuthHostname + '\'' +
                    ", p4oAuthPort=" + p4oAuthPort +
                    ", login='" + login + '\'' +
                    ", redirectUri='" + redirectUri + '\'' +
                    '}';
        }
    }
}
